/*=====================================================================*/
/*    serrano/prgm/project/bigloo/runtime/Clib/cports.c                */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Thu Jul 23 15:34:53 1992                          */
/*    Last change :  Tue Jun 21 09:03:08 2005 (serrano)                */
/*    -------------------------------------------------------------    */
/*    Input ports handling                                             */
/*=====================================================================*/
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef _MSC_VER
#   include <dirent.h>
#endif
#include <string.h>
#if( !(defined( NeXT ) && (defined( mc68000 ) || defined( i386 ))) )
#   ifdef HAVE_TERMIO
#      include <termio.h>
#   endif
#endif
#if !defined( sony_news ) && \
    !(defined( NeXT ) && (defined( mc68000 ) || defined( i386 ))) && \
    !defined( _MSC_VER )
#   include <unistd.h>
#endif
#ifndef _MSC_VER
#  include <sys/file.h>
#endif
#include <bigloo.h>
#if( defined( sony_news ) || (defined( NeXT ) && defined( mc68000 )) )
#   include <ctype.h>
#endif

#if POSIX_FILE_OPS
#   include <unistd.h>
#endif

#ifdef _MSC_VER
#   include <io.h>
#   include <windows.h>
#   define pclose _pclose
#   define popen _popen
#   define S_ISDIR( mode ) ((mode & _S_IFDIR) != 0)
#endif

/*---------------------------------------------------------------------*/
/*    isascii                                                          */
/*---------------------------------------------------------------------*/
#if( !defined( isascii ) )
#   define isascii( c ) (!((c) & ~0177))
#endif

/*---------------------------------------------------------------------*/
/*    Global variables                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF long default_io_bufsiz;

/*---------------------------------------------------------------------*/
/*    External definitions.                                            */
/*---------------------------------------------------------------------*/
extern obj_t string_to_bstring();
extern obj_t string_to_keyword();
extern obj_t make_input_port( char *, FILE *, obj_t, long );
extern obj_t make_string();

#if( BGL_HAVE_SENDFILE )
extern ssize_t sendfile( int, int, off_t *, size_t );
#endif

/*---------------------------------------------------------------------*/
/*    Prototypes                                                       */
/*---------------------------------------------------------------------*/
static bool_t pipe_name_p( char * );
static char *pipe_name( char * );

static size_t strwrite( void *, size_t, size_t, void *stream );
static int strputc( int, void * );

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    bgl_feof ...                                                     */
/*    -------------------------------------------------------------    */
/*    We explicitly define this function because on some system        */
/*    feof is a macro.                                                 */
/*---------------------------------------------------------------------*/
int
bgl_feof( FILE *stream ) {
   return feof( stream );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    proc_eof ...                                                     */
/*---------------------------------------------------------------------*/
static int
proc_eof( FILE *port ) {
   CREF( (obj_t)port )->input_port_t.eof = 1;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    console_fread ...                                                */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input, if */
/*    not enough characters are available.                             */
/*    -------------------------------------------------------------    */
/*    This function is prefered to bgl_pipe_fread for consoles because */
/*    it automatically flushes stdout before reading (see getc).       */
/*---------------------------------------------------------------------*/
static int
console_fread( char *ptr, long size, long nmemb, FILE *stream ) {
   long  num = size * nmemb;
   char *buf = ptr;
   int   c;

   while( ((c = getc( stream )) != EOF) ) {
      *buf++ = c;

      if( c == '\n' ) break;
      if( --num <= 0 ) break;
   }

   return (long)(buf - ptr);
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    bgl_pipe_fread ...                                               */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input if  */
/*    insufficient characters are available.                           */
/*---------------------------------------------------------------------*/
int
bgl_pipe_fread( char *ptr, long size, long nmemb, FILE *stream ) {
#ifdef POSIX_FILE_OPS
   long  num = size * nmemb;

   return read( fileno( stream ), ptr, num );
#else
   return console_fread( ptr, size, nmemb, stream );
#endif
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    proc_read ...                                                    */
/*---------------------------------------------------------------------*/
static int
proc_read( char *b, int s, int n, obj_t port ) {
   obj_t proc = CREF( port )->input_port_t.name;
   obj_t rest = CREF( port )->input_port_t.userdata;
   obj_t o = STRINGP( rest ) ? rest : PROCEDURE_ENTRY( proc )( proc, BEOA );

   if( o == BFALSE ) {
      CREF( port )->input_port_t.eof = 1;
      return 0;
   } else {
      if( STRINGP( o ) ) {
	 int l = s * n;
	 int r = STRING_LENGTH( o );
	 char *s = BSTRING_TO_STRING( o );

	 if( r <= l ) {
	    memmove( b, s, r );
	    
	    CREF( port )->input_port_t.userdata = BUNSPEC;
	    
	    return r;
	 } else {
	    memmove( b, s, l );
	    
	    CREF( port )->input_port_t.userdata =
	       string_to_bstring_len( BSTRING_TO_STRING( &s[ l ] ), r - l );
				      
	    return l;
	 }
      } else {
	 C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			   "input-procedure-port",
			   "Procedure result must be a #f, or a string",
			   o );
      }
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    fileflush ...                                                    */
/*---------------------------------------------------------------------*/
static obj_t
fileflush( obj_t o ) {
   return ((fflush( OUTPUT_PORT( o ).ostream ) == EOF) ? BFALSE : BTRUE);
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    strflush ...                                                     */
/*---------------------------------------------------------------------*/
static obj_t
strflush( obj_t port ) {
   obj_t res;

   if( OUTPUT_STRING_PORT(port).buffer ) {
      OUTPUT_STRING_PORT(port).buffer[ OUTPUT_STRING_PORT(port).offset ] = 0;
      res = string_to_bstring( OUTPUT_STRING_PORT(port).buffer );
/*       OUTPUT_STRING_PORT(port).buffer[ 0 ] = 0;                     */
/*       OUTPUT_STRING_PORT(port).offset = 0;                          */
      
      return res;
   } else {
      return string_to_bstring( "" );
   }
}

/*---------------------------------------------------------------------*/
/*     make_output_port ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
make_output_port( char *name, FILE *file, obj_t kindof ) {
   obj_t new_output_port;

   new_output_port = GC_MALLOC( OUTPUT_PORT_SIZE );

   new_output_port->output_port_t.header = MAKE_HEADER( OUTPUT_PORT_TYPE, 0 );
   new_output_port->output_port_t.ostream = file;
   new_output_port->output_port_t.name = string_to_bstring( name );
   new_output_port->output_port_t.kindof = kindof;
   new_output_port->output_port_t.sysputc = fputc;
   new_output_port->output_port_t.syswrite = fwrite;
   new_output_port->output_port_t.sysflush = fileflush;
   new_output_port->output_port_t.chook = BUNSPEC;
   new_output_port->output_port_t.userdata = BUNSPEC;

   switch( (long)kindof ) {
      case (long)KINDOF_FILE:
      case (long)KINDOF_PROCPIPE:
	 new_output_port->output_port_t.sysclose = fclose;
	 break;
	 
      case (long)KINDOF_PIPE:
	 new_output_port->output_port_t.sysclose = pclose;
	 break;

      case (long)KINDOF_CONSOLE:
	 new_output_port->output_port_t.sysclose = 0L;
	 break;

      default:
	 fprintf( stderr, "ERROR: make_output_port\n" );
	 new_output_port->output_port_t.sysclose = 0L;
   }

   return BREF( new_output_port );
}

/*---------------------------------------------------------------------*/
/*    open_output_file ...                                             */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
open_output_file( obj_t name ) {
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      if( !(file = popen( pipe_name( cname ), "w" )) )
         return BFALSE;

       return make_output_port( cname, file, KINDOF_PIPE );
   } else
#endif
   {
      if (strcmp( cname, "null:" ) == 0)
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif

      if( !(file = fopen( cname, "wb" )) )
         return BFALSE;

      return make_output_port( cname, file, KINDOF_FILE );
   }
}

/*---------------------------------------------------------------------*/
/*    append_output_file ...                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
append_output_file( obj_t name ) {
   FILE *file;

   if( !(file = fopen( BSTRING_TO_STRING(name), "a+b ")) )
      return BFALSE;

   return make_output_port( BSTRING_TO_STRING( name ),
			    file,
			    KINDOF_FILE );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_output_string ...                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
open_output_string() {
   char *buffer;
   obj_t port;

   port = GC_MALLOC( OUTPUT_STRING_PORT_SIZE );

   /* We allocate a buffer filled of zero */
   buffer = (char *)(GC_MALLOC_ATOMIC(OUTPUT_STRING_PORT_BUFFER_SIZE + 1));
   bzero( buffer, OUTPUT_STRING_PORT_BUFFER_SIZE + 1 );
   
   port->output_port_t.header = MAKE_HEADER(OUTPUT_STRING_PORT_TYPE, 0);
   port->output_port_t.kindof = KINDOF_STRING;
   port->output_port_t.sysputc = strputc;
   port->output_port_t.syswrite = strwrite;
   port->output_port_t.sysclose = 0L;
   port->output_port_t.sysflush = strflush;
   port->output_port_t.ostream = port;
   port->output_port_t.chook = BUNSPEC;
   port->output_port_t.userdata = BUNSPEC;
   
   port->output_string_port_t.buffer = buffer;
   port->output_string_port_t.size = OUTPUT_STRING_PORT_BUFFER_SIZE;
   port->output_string_port_t.offset = 0;

   return BREF( port );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    get_output_string ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
get_output_string( obj_t port ) {
   if( OUTPUT_STRING_PORTP( port ) ) {
      if( OUTPUT_STRING_PORT (port ).size == 0 )
	 return make_string( 0, ' ' );
      else
	 return string_to_bstring_len( OUTPUT_STRING_PORT(port).buffer,
				       OUTPUT_STRING_PORT(port).offset);
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"get-output-string",
			"Not a string port",
			port );
      return BUNSPEC;
   }
}

/*---------------------------------------------------------------------*/
/*    close_output_port ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
close_output_port( obj_t port ) {
   obj_t res = port; 
   if( OUTPUT_PORTP( port ) ) {
      obj_t chook = OUTPUT_PORT_CHOOK( port );
      
      if( OUTPUT_STRING_PORTP( port ) ) {
	 res = get_output_string( port );
      }

      if( OUTPUT_PORT( port ).sysclose ) 
	 OUTPUT_PORT( port ).sysclose( OUTPUT_PORT( port ).ostream );
      
      OUTPUT_PORT(port).kindof = KINDOF_CLOSED;
      OUTPUT_PORT( port ).sysclose = 0L;
      
      if( PROCEDUREP( chook ) ) {
	 if( PROCEDURE_ARITY( chook ) == 1 ) {
	    PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "close-output-port",
			      "illegal close hook arity",
			      chook );
	 }
      }
   }
   
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    make_input_port ...                                              */
/*---------------------------------------------------------------------*/
obj_t
make_input_port( char *name, FILE * file, obj_t kindof, long bufsiz ) {
   obj_t new_input_port;

   /* An input port cannot be allocated as an atomic object    */
   /* because it holds a buffer and a name that are GC objects */
   new_input_port = GC_MALLOC( INPUT_PORT_SIZE );

   new_input_port->input_port_t.header = MAKE_HEADER( INPUT_PORT_TYPE, 0 );
   new_input_port->input_port_t.kindof = kindof;
   new_input_port->input_port_t.name = string_to_bstring( name );
   new_input_port->input_port_t.file = file;
   new_input_port->input_port_t.filepos = 0;
   new_input_port->input_port_t.fillbarrier = -1;
   new_input_port->input_port_t.bufsiz = bufsiz;
   new_input_port->input_port_t.eof = 0;
   new_input_port->input_port_t.matchstart = 0;
   new_input_port->input_port_t.matchstop = 0;
   new_input_port->input_port_t.forward = 0;
   new_input_port->input_port_t.bufpos = 1;
   new_input_port->input_port_t.lastchar = '\n';
   new_input_port->input_port_t.syseof = bgl_feof;
   new_input_port->input_port_t.chook = BUNSPEC;
   new_input_port->input_port_t.userdata = BUNSPEC;

   switch( (long)kindof ) {
     case (long)KINDOF_CONSOLE:
	new_input_port->input_port_t.sysclose = 0;
	new_input_port->input_port_t.sysread = console_fread;
	break;
	
#if( HAVE_PIPE )
     case (long)KINDOF_PIPE:
	new_input_port->input_port_t.sysclose = pclose;
	new_input_port->input_port_t.sysread = bgl_pipe_fread;
	break;
	
     case (long)KINDOF_PROCPIPE:
#endif	
     case (long)KINDOF_SOCKET:
	new_input_port->input_port_t.sysclose = fclose;
	new_input_port->input_port_t.sysread = bgl_pipe_fread;
	break;

     case (long)KINDOF_FILE:
	new_input_port->input_port_t.sysclose = fclose;
	new_input_port->input_port_t.sysread = (int (*)())fread;
	break;

     case (long)KINDOF_PROCEDURE:
	new_input_port->input_port_t.syseof = proc_eof;
	new_input_port->input_port_t.sysclose = 0;
	new_input_port->input_port_t.sysread = proc_read;
	break;

     default:
	new_input_port->input_port_t.sysclose = 0;
	new_input_port->input_port_t.sysread = (int (*)())fread;
     }


   if( bufsiz > 0 ) {
      new_input_port->input_port_t.buffer = 
	 (unsigned char *)GC_MALLOC_ATOMIC( bufsiz + 1 );
      new_input_port->input_port_t.buffer[ 0 ] = '\0';
   } else
      new_input_port->input_port_t.buffer = '\0';

   return BREF( new_input_port );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_input_pipe ...                                              */
/*---------------------------------------------------------------------*/
obj_t
open_input_pipe( obj_t name, obj_t bbufsiz ) {
#if( HAVE_PIPE )
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

   if( !(file = popen( cname, "r" )) )
      return BFALSE;

   /* we use our own buffer */
   setvbuf( file, NULL, _IONBF, 0 );  

   return make_input_port( cname, file, KINDOF_PIPE, CINT( bbufsiz ) );
#else
   return BFLASE;
#endif
}

/*---------------------------------------------------------------------*/
/*    open_input_file ...                                              */
/*    -------------------------------------------------------------    */
/*    !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!!  */
/*    We fill up its associated buffer when opening an input port.     */
/*    -------------------------------------------------------------    */
/*    This function open two kind of files. Regular file and Unix      */
/*    like pipes when the file name is something like "| ...".         */
/*---------------------------------------------------------------------*/
obj_t
open_input_file( obj_t name, obj_t bbufsiz ) {
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      if( !(file = popen( pipe_name( cname ), "r" ) ) )
	 return BFALSE;

      /* we use our own buffer */
      setvbuf( file, NULL, _IONBF, 0 );  

      return make_input_port( cname, file, KINDOF_PIPE,
			      CINT( bbufsiz ) );
   } else
#endif
   {
      if (strcmp( cname, "null:" ) == 0)
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif

      if( !(file = fopen( cname, "rb" )) )
         return BFALSE;

      /* we use our own buffer */
      setvbuf( file, NULL, _IONBF, 0 );  

      return make_input_port( cname, file, KINDOF_FILE, CINT( bbufsiz ) );
   }
}

/*---------------------------------------------------------------------*/
/*    open_input_console ...                                           */
/*---------------------------------------------------------------------*/
obj_t
open_input_console() {
   return make_input_port("[stdin]", stdin, KINDOF_CONSOLE, default_io_bufsiz);
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    file_to_buffered_input_port ...                                  */
/*---------------------------------------------------------------------*/
obj_t
file_to_buffered_input_port( FILE * file, long bufsize ) {
   if( file == stdin )
      return open_input_console();
   else {
      long bsize = bufsize <= 0 ? default_io_bufsiz : bufsize;

      return make_input_port( "[file]", file, KINDOF_FILE, bsize );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    file_to_input_port ...                                           */
/*---------------------------------------------------------------------*/
obj_t
file_to_input_port( FILE * file ) {
   return file_to_buffered_input_port( file, -1 );
}

/*---------------------------------------------------------------------*/
/*    open_input_string ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
open_input_string( obj_t string ) {
   obj_t port;
   long bufsiz = STRING_LENGTH( string );

   port = make_input_port( "[string]", 0L, KINDOF_STRING, bufsiz + 1 );

   CREF(port)->input_port_t.eof = 1;
   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   memcpy( &RGC_BUFFER(port)[ 0 ], BSTRING_TO_STRING( string ), bufsiz );
   RGC_BUFFER(port)[ bufsiz ] = '\0';

   return port;
}

/*---------------------------------------------------------------------*/
/*    open_input_procedure ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
open_input_procedure( obj_t fun, obj_t bfz ) {
   if( PROCEDURE_CORRECT_ARITYP( fun, 0 ) ) {
      obj_t port = make_input_port( "", 0L, KINDOF_PROCEDURE, CINT( bfz )+1 );

      CREF( port )->input_port_t.file = port;
      CREF( port )->input_port_t.name = fun;
      CREF( port )->input_port_t.userdata = BUNSPEC;

      return port;
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"open-input-procedure",
			"Illegal procedure arity",
			fun );
   }
}


/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_input_c_string ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
open_input_c_string( char *c_string ) {
   obj_t port;
   long bufsiz = (long)strlen( c_string );
   char *new_string = 0;

   port = make_input_port( "[c_string]", 0L, KINDOF_STRING, 0 );

   CREF(port)->input_port_t.eof = 1;
   CREF(port)->input_port_t.bufsiz = bufsiz + 1;
   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   CREF(port)->input_port_t.buffer = (unsigned char *)c_string;

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    reopen_input_c_string ...                                        */
/*    -------------------------------------------------------------    */
/*    Simply changes the input buffer of an input string. Does not     */
/*    re-allocate a brand new input-port.                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
reopen_input_c_string( obj_t port, char *c_string ) {
   long bufsiz = (long)strlen(c_string);

   if( CREF(port)->input_port_t.bufsiz < (bufsiz + 1) ) {
      CREF(port)->input_port_t.bufsiz = bufsiz + 1;
      CREF(port)->input_port_t.buffer =
	 (unsigned char *)GC_MALLOC_ATOMIC(bufsiz + 1);
   }

   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   CREF(port)->input_port_t.matchstart = 0;
   CREF(port)->input_port_t.matchstop = 0;
   CREF(port)->input_port_t.forward = 0;
   CREF(port)->input_port_t.lastchar = '\n';
   strcpy( (char *)(CREF(port)->input_port_t.buffer), (char *)c_string );

   return port;
}

/*---------------------------------------------------------------------*/
/*    close_input_port ...                                             */
/*    -------------------------------------------------------------    */
/*    This function does erase the RGC_BUFFER otherwise we could       */
/*    have a problem in multithreaded environment (a thread reading    */
/*    while another one is closing the same port).                     */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
close_input_port( obj_t port ) {
   if( INPUT_PORTP( port ) ) {
      if( (INPUT_PORT( port ).kindof != KINDOF_CLOSED) &&
	  (INPUT_PORT( port ).kindof != KINDOF_CONSOLE) ) {
	 obj_t chook = INPUT_PORT_CHOOK( port );
	 
	 if( INPUT_PORT( port ).sysclose ) 
	    INPUT_PORT( port ).sysclose( INPUT_PORT( port ).file );
	 
	 INPUT_PORT( port ).eof = 1;
      	 INPUT_PORT( port ).kindof = KINDOF_CLOSED;
	 INPUT_PORT( port ).sysclose = 0L;
	 
	 if( PROCEDUREP( chook ) ) {
	    if( PROCEDURE_ARITY( chook ) == 1 ) {
	       PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	    } else {
	       C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
				 "close-input-port",
				 "illegal close hook arity",
				 chook );
	    }
	 }
      }

   }

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_seek ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
bgl_input_port_seek( obj_t port, long pos ) {
   if( INPUT_PORT_ON_FILEP( port ) ) {
      if( fseek( INPUT_PORT( port ).file, pos, SEEK_SET ) )
	 return BFALSE;
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).eof = 0;
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).forward = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).lastchar = '\n';
      INPUT_PORT( port ).buffer[0] = '\0';

      return BTRUE;
   }

   if( INPUT_PORT_ON_STRINGP( port ) && (pos < INPUT_PORT( port ).bufsiz) ) {
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).matchstart = pos;
      INPUT_PORT( port ).matchstop = pos;
      INPUT_PORT( port ).forward = pos;

      return BTRUE;
   }

   return BFALSE;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_output_port_seek ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
bgl_output_port_seek( obj_t port, long pos ) {
   if( OUTPUT_FILE_PORTP( port ) ) {
      if( fseek( OUTPUT_PORT(port).ostream, pos, SEEK_SET ) )
	 return BFALSE;
      return BTRUE;
   }
   if( OUTPUT_STRING_PORTP( port ) ) {
      if( pos >= OUTPUT_STRING_PORT( port ).size )
	 return BFALSE;

      OUTPUT_STRING_PORT( port ).offset = pos;
      return BTRUE;
   }
   return BFALSE;
}
   
/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_reopen ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
bgl_input_port_reopen( obj_t port ) {
   FILE *nf;
   
   if( !INPUT_PORT_ON_FILEP( port ) )
      return BFALSE;

   nf = freopen( BSTRING_TO_STRING( INPUT_PORT( port ).name ),
		 "r",
		 (FILE *)INPUT_PORT(port).file );

   if( !nf ) return BFALSE;
   
   INPUT_PORT( port ).file = (void *)nf;
      
   /* we use our own buffer */
   setvbuf( (FILE *)INPUT_PORT( port ).file, NULL, _IONBF, 0 );

   INPUT_PORT( port ).filepos = 0;
   INPUT_PORT( port ).eof = 0;
   INPUT_PORT( port ).matchstart = 0;
   INPUT_PORT( port ).matchstop = 0;
   INPUT_PORT( port ).forward = 0;
   INPUT_PORT( port ).bufpos = 1;
   INPUT_PORT( port ).lastchar = '\n';
   INPUT_PORT( port ).buffer[ 0 ] = '\0';

   return BTRUE;
}

/*---------------------------------------------------------------------*/
/*    obj                                                              */
/*    reset_console ...                                                */
/*    -------------------------------------------------------------    */
/*    We flush input port, for ^C to work correctly within the         */
/*    interpreter. The only place this function is called is in the    */
/*    REPL (see Eva/eval.scm).                                         */
/*---------------------------------------------------------------------*/
obj_t
reset_console( obj_t port ) {
   if( INPUT_PORT( port ).kindof == KINDOF_CONSOLE ) {
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).buffer[ 0 ] = '\0';
      INPUT_PORT( port ).lastchar = '\n';
   }

   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*     bgl_init_io ...                                                 */
/*---------------------------------------------------------------------*/
void
bgl_init_io() {
#if( !defined( _SBFSIZ ) )
#   define _SBFSIZ 1
#endif

   default_io_bufsiz = BUFSIZ * _SBFSIZ;

   BGL_CURRENT_OUTPUT_PORT_SET(
      make_output_port( "stdout", stdout, KINDOF_CONSOLE ) );
   BGL_CURRENT_ERROR_PORT_SET(
      make_output_port( "stderr", stderr, KINDOF_CONSOLE ) );
   BGL_CURRENT_INPUT_PORT_SET(
      open_input_console() );
}

/*---------------------------------------------------------------------*/
/*    fexists ...                                                      */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
bool_t
fexists( char *name ) {
#if( HAVE_PIPE )
   if( pipe_name_p( name ) ) {
      return 1;
   }
#else
   if( pipe_name_p(name) ) {
      return 0;
   }
#endif

# ifndef _MSC_VER
    return !access( name, F_OK );
# else
    /* !!!!! verify semantics of Unix' access */
    return !_access( name, 0 );        
# endif
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    reset_eof ...                                                    */
/*    -------------------------------------------------------------    */
/*    The function erase the end-of-file of input console port.        */
/*    This allows, restart reading after a ^D.                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
bool_t
reset_eof( obj_t port ) {
   if( INPUT_PORT( port ).kindof == KINDOF_CONSOLE ) {
      /* we forget about EOF */
      INPUT_PORT( port ).eof = 0;

      /* we cleanup buffer   */
      reset_console( port );

      /* we clear errors.    */
      clearerr( stdin );

      return 1;
   }
   else
      return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_strport_grow ...                                             */
/*---------------------------------------------------------------------*/
obj_t
bgl_strport_grow( obj_t p ) {
   char *buffer = OUTPUT_STRING_PORT(p).buffer;
   long old_size = OUTPUT_STRING_PORT(p).size;
   long new_size;

   if( old_size == 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR, "string-port", "port closed", p );

      return p;
   } else {
      new_size = old_size * 2;
      buffer = (char *)GC_realloc( (obj_t)buffer, new_size + 1 );
      buffer[ old_size ] ='\0';
      OUTPUT_STRING_PORT(p).buffer = buffer;

      OUTPUT_STRING_PORT(p).size = new_size;

      return p;
   }
}

/*---------------------------------------------------------------------*/
/*    static bool_t                                                    */
/*    pipe_name_p ...                                                  */
/*    -------------------------------------------------------------    */
/*    Is a file name a pipe name ? A pipe name start by the            */
/*    sequence "| ".                                                   */
/*---------------------------------------------------------------------*/
static bool_t
pipe_name_p( char *name ) {
   return( (name[ 0 ] == '|') && (name[ 1 ] == ' ') );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    pipe_name ...                                                    */
/*    -------------------------------------------------------------    */
/*    Pipe name to name translation.                                   */
/*---------------------------------------------------------------------*/
static char *
pipe_name( char *pipe_name ) {
   return (pipe_name + 1);
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    directoryp ...                                                   */
/*    -------------------------------------------------------------    */
/*    Is a file a directory?                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
bool_t
directoryp( char *name ) { 
   struct stat buf;

   if( stat( name, &buf ) == -1 )
      return 0;

   return S_ISDIR( buf.st_mode & S_IFMT );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    directory_to_list ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF
obj_t
directory_to_list( char *name ) {
   obj_t res = BNIL;
#ifndef _MSC_VER
   DIR *dir;
   struct dirent *dirent;

   if( (dir = opendir( name )) ) {
      while( (dirent = readdir( dir )) ) {
	 char *fname = dirent->d_name;

	 if( strcmp( fname, "." ) && strcmp( fname, ".." ) )
	    res = MAKE_PAIR( string_to_bstring( fname ), res );
      }
      closedir( dir );
   }
#else
   char *const path = (char *)malloc( strlen( name ) + 2 + 1 );

   strcpy( path, name );
   strcat( path, "\\*" );

   {
      WIN32_FIND_DATA find_data;
      HANDLE hSearch = FindFirstFile( path, &find_data );

      if( hSearch != INVALID_HANDLE_VALUE ) {
         BOOL keep_going;

         do {
            if( (strcmp( find_data.cFileName, "." ) != 0)
                && (strcmp( find_data.cFileName, ".." ) != 0) )
               res = MAKE_PAIR( string_to_bstring( find_data.cFileName ),
				res );
            keep_going= FindNextFile( hSearch, &find_data );
         } while( keep_going );

         FindClose( hSearch );
      }
   }
#endif
   return res;
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    copyfile ...                                                     */
/*    -------------------------------------------------------------    */
/*    This function is a replacement for sendfile. In particular       */
/*    it is used when the INPUT-PORT is a socket descriptor or when    */
/*    the host system does not support the sendfile system call.       */
/*---------------------------------------------------------------------*/
static int
copyfile( int on, int in, int sz ) {
#ifdef __GNUC__
   char buf[ default_io_bufsiz + 1 ];
#else
   char *buf = alloca( default_io_bufsiz + 1 );
#endif
   int rsz = 0;

   if( sz < 0 ) {
      int n;
      while( (n = read( in, buf, default_io_bufsiz )) > 0 ) {
	 write( on, buf, n );
	 rsz += n;
      }

      return rsz;
   } else {
      int n;
      int s = (default_io_bufsiz > sz) ? sz : default_io_bufsiz;

      while( (s > 0) && (n = read( in, buf, s )) > 0 ) {
	 write( on, buf, n );
	 rsz += n;
	 sz -= n;
	 s = (default_io_bufsiz > sz) ? sz : default_io_bufsiz;
      }

      return rsz;
   }
}   
   
/*---------------------------------------------------------------------*/
/*    bgl_sendchars ...                                                */
/*    -------------------------------------------------------------    */
/*    uses sendfile to "copy" the input-port to the output-port        */
/*    flushes output-port!                                             */
/*---------------------------------------------------------------------*/
obj_t
bgl_sendchars( obj_t ip, obj_t op, int sz ) {
#define inp INPUT_PORT( ip )
#define outp OUTPUT_PORT( op )
   int dsz;
   int ws = 0;
   struct stat in;
   struct stat out;
   int n;

   switch( (long)(inp.kindof) ) {
      case (long)KINDOF_STRING:
      case (long)KINDOF_CLOSED:
      case (long)KINDOF_PROCEDURE:
	 return BFALSE;
   }

   if( OUTPUT_STRING_PORTP( op ) )
       return BFALSE;
       
   switch( (long)(outp.kindof) ) {
      case (long)KINDOF_CLOSED:
      case (long)KINDOF_PROCEDURE:
	 return BFALSE;
   }

   dsz = inp.bufpos - inp.matchstop - 1;
   
   fflush( (FILE *)outp.ostream );

   if( dsz > 0 ) {
      /* flush the buffer when it contains characters */
      ws = ((sz > 0) && (dsz > sz)) ? sz : dsz;
      fwrite( &(inp.buffer[ inp.matchstop ]), ws, 1, (FILE *)outp.ostream );

      inp.matchstop += ws;
      inp.forward = inp.matchstop;
      
      if( ferror( (FILE *)outp.ostream ) || fflush( (FILE *)outp.ostream ) ) {
	 C_SYSTEM_FAILURE( BGL_IO_ERROR, "sendchars", strerror( errno ), op );
	 return BINT( 0 );
      }

      if( (sz > 0) && (dsz > sz) ) {
	 return BINT( ws );
      }

      sz -= ws;
   }
   
#if( BGL_HAVE_SENDFILE )
   if( 
      /* Some operating systems (such as Linux 2.6.10) are demanding */
      /* on the input and output port. These requirements are set    */
      /* in the configuration files and used to determine what as    */
      /* to be checked before invoking the actual sendfile sys call  */
#  if( BGL_SENDFILE_REQUIRE_INPUT_FILE )      
       (inp.kindof != KINDOF_FILE)
       || fstat( fileno( (FILE *)inp.file ), &in )
       || !S_ISREG( in.st_mode )
#  endif
#  if( BGL_SENDFILE_REQUIRE_OUTPUT_SOCKET )
       || (outp.kindof != KINDOF_FILE)
       || fstat( fileno( (FILE *)outp.ostream ), &out )
       || ((out.st_mode & S_IFSOCK) != S_IFSOCK)
#  endif       
       || 0 ) {
#endif
      n = copyfile( fileno( (FILE *)outp.ostream ),
		    fileno( (FILE *)inp.file ),
		    sz );
#if( BGL_HAVE_SENDFILE )
   } else {
      if( sz == -1 ) sz = in.st_size;

      bgl_gc_start_blocking();
      n = sendfile( fileno( (FILE *)outp.ostream ),
		    fileno( (FILE*)inp.file ),
		    0,
		    sz );
      bgl_gc_stop_blocking();
   }
#endif
   if( n == -1 ) {
      C_SYSTEM_FAILURE( BGL_IO_ERROR,
			"sendchars",
			strerror( errno ),
			MAKE_PAIR( ip, op ) );
      return BINT( 0 );
   }
     
   inp.filepos += ws + n;
   fseek( (FILE *)inp.file, inp.filepos, SEEK_SET );

   return BINT( ws + n );
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    strwrite ...                                                     */
/*---------------------------------------------------------------------*/
static size_t
strwrite( void *ptr, size_t size, size_t nmemb, void *stream ) {
   long len = (long)(size == 1 ? nmemb : size * nmemb);
   obj_t p = (obj_t)stream;
   long offset = OUTPUT_STRING_PORT( p ).offset;

   while( (OUTPUT_STRING_PORT( p ).offset+len) > OUTPUT_STRING_PORT( p ).size )
      bgl_strport_grow( p );

   memcpy( &(OUTPUT_STRING_PORT(p).buffer[ offset ] ), ptr, len );

   OUTPUT_STRING_PORT( p ).offset = offset + len;

   return len;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    strputc ...                                                      */
/*---------------------------------------------------------------------*/
static int
strputc( int c, void *stream ) {
   obj_t p = (obj_t)stream;
   long offset = OUTPUT_STRING_PORT( p ).offset;
   
   if( END_OF_STRING_PORTP( p ) )
      bgl_strport_grow( p );

   OUTPUT_STRING_PORT( p ).buffer[ offset ] = c;
   OUTPUT_STRING_PORT( p ).offset = offset + 1;

   return c;
}
 
